"""
Module containing generic utility functions and classes.

This module provides utility functions and classes for logging and other purposes.
"""
from __future__ import annotations
from typing import TYPE_CHECKING, NoReturn

import argparse
import logging
import logging.handlers
import re
import collections
from datetime import datetime, timezone

if TYPE_CHECKING:
    from typing import Dict, Tuple, Any, MutableSequence, Optional


def robust_time_parse(time_string: str) -> datetime:
    """
    This function replaces the 'Z' character with '+00:00' to handle UTC time and ensures that fractional
    seconds are padded to six digits if necessary.

    Args:
        time_string (str): The time string to be parsed. Expected format is 'YYYY-MM-DDTHH:MM:SS.ssssss+HH:MM'
                          or 'YYYY-MM-DDTHH:MM:SS.ssssssZ'.

    Returns:
        datetime: A datetime object representing the parsed time string.

    Raises:
        ValueError: If the time string is not in a valid ISO 8601 format.
    """
    time_string = time_string.replace('Z', '+00:00')
    match: re.Match[str] | None = re.search(
        r'^(?P<start>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.)(?P<fractions>\d+)(?P<end>\+\d{2}:\d{2})$', time_string)
    if match:
        time_string = match.group('start') + match.group('fractions').ljust(6, "0") + match.group('end')
    return datetime.fromisoformat(time_string)


def log_extra_keys(log: logging.Logger, where: str, dictionary: Dict[str, Any], allowed_keys: Optional[set[str]] = None) -> None:
    """
    Logs a warning if there are any keys in the dictionary that are not in the allowed_keys set.

    Args:
        log (logging.log): The logger instance to use for logging warnings.
        log (logging.Logger): The logger instance to use for logging warnings.
        dictionary (Dict[str, Any]): The dictionary to check for extra keys.
        allowed_keys (set[str]): The set of keys that are allowed in the dictionary.

    Returns:
        None
    """
    if allowed_keys is None:
        allowed_keys = set()
    extra_keys = set(dictionary.keys()) - allowed_keys
    if extra_keys:
        log.info(f"Unexpected keys found in {where}: {extra_keys} Dictionary is {dictionary}")


def config_remove_credentials(config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Removes any credentials from the configuration dictionary.

    Args:
        config (Dict[str, Any]): The configuration dictionary to remove credentials from.

    Returns:
        Dict[str, Any]: The configuration dictionary with credentials removed.
    """
    def __recursive_remove_credentials(config: Dict[str, Any]) -> Dict[str, Any]:
        for key in config:
            if isinstance(config[key], dict):
                config[key] = __recursive_remove_credentials(config[key])
            if isinstance(config[key], list):
                for i, item in enumerate(config[key]):
                    if isinstance(item, dict):
                        config[key][i] = __recursive_remove_credentials(item)
            if 'pin' in key.lower() or 'secret' in key.lower() or 'password' in key.lower() or 'token' in key.lower():
                config[key] = '***'
        return config
    config_copy: dict[str, Any] = config.copy()
    __recursive_remove_credentials(config_copy)
    return config_copy


# pylint: disable=too-few-public-methods
class DuplicateFilter(logging.Filter):
    """
    A logging filter that suppresses duplicate log messages from the same module within a specified time frame.
    """

    def __init__(self, do_not_filter_above=logging.ERROR, filter_reset_seconds: int = 0, name: str = '') -> None:
        super().__init__(name=name)
        self._last_log: Dict[str, Dict[int, Tuple[Tuple[str, Any], datetime]]] = {}
        self._first_time: bool = True
        self.do_not_filter_above: int = do_not_filter_above
        self.filter_reset_seconds: int = filter_reset_seconds

    def filter(self, record: logging.LogRecord) -> bool:
        # don't filter messages above the specified level
        if record.levelno >= self.do_not_filter_above:
            return True

        now: datetime = datetime.now(tz=timezone.utc)

        # were messages from this module already logged?
        if record.module in self._last_log:
            time_since_last_log = (now - self._last_log[record.module][record.levelno][1]).total_seconds()
            if record.levelno in self._last_log[record.module]:
                # were the same message and arguments logged?
                if self._last_log[record.module][record.levelno][0] == (record.msg, record.args):
                    # was the message logged within the specified time frame?
                    if time_since_last_log < self.filter_reset_seconds:
                        # inform user about repeated messages being suppressed only the first time
                        if self._first_time:
                            self._first_time = False
                            logging.info('Repeated log messages from the same module are hidden (does not apply to errors or critical problems)')
                        # indicate that the message should be suppressed
                        return False
            # store the message and the time it was logged
            self._last_log[record.module][record.levelno] = ((record.msg, record.args), now)
        else:
            # store the message and the time it was logged
            self._last_log[record.module] = {record.levelno: ((record.msg, record.args), now)}
        return True


class ThrowingArgumentParser(argparse.ArgumentParser):
    """
    A custom argument parser that raises an exception instead of exiting on error.

    This class extends `argparse.ArgumentParser` and overrides the `error` method
    to raise an `argparse.ArgumentError` instead of printing an error message and
    exiting the program.

    Methods:
        error(message: str) -> NoReturn:
            Raises an `argparse.ArgumentError` with the provided error message.

    Args:
        message (str): The error message to be included in the raised exception.
    """
    def error(self, message) -> NoReturn:
        raise argparse.ArgumentError(argument=None, message=message)


class LogMemoryHandler(logging.handlers.MemoryHandler):  # pylint: disable=protected-access
    """
    A custom memory handler that can be used to store log records in memory.

    This class extends `logging.handlers.MemoryHandler` and overrides the `flush`
    method to return the stored log records instead of writing them to a target.
    """
    def __init__(self, capacity: int = 1000) -> None:
        self._capacity: int = capacity
        self.storage: MutableSequence[logging.LogRecord] = collections.deque(maxlen=capacity)
        super().__init__(capacity, flushLevel=logging.ERROR, target=None)

    @property
    def capacity(self) -> int:
        """
        Returns the capacity of the memory handler.

        Returns:
            int: The capacity of the memory handler.
        """
        return self._capacity

    @capacity.setter
    def capacity(self, new_capacity: int) -> None:
        """
        Sets the capacity of the memory handler.

        Args:
            new_capacity (int): The new capacity of the memory handler.
        """
        self.storage = collections.deque(self.storage, maxlen=new_capacity)

    def emit(self, record: logging.LogRecord) -> None:
        self.storage.append(record)

    def flush(self) -> None:
        """
        Flush stored log records.
        Does nothing in this implementation.
        """
        return



"""
This module defines general enums.
"""

from enum import Enum


class ConnectionState(Enum):
    """
    Enum for conenction state
    """
    DISCONNECTED = 'disconnected'
    CONNECTING = 'connecting'
    CONNECTED = 'connected'
    DISCONNECTING = 'disconnecting'
    ERROR = 'error'
    UNKNOWN = 'unknown connection state'



"""
This module defines enumerations for various units of measurement, including temperature and length.

Classes:
    Temperature (Enum): An enumeration representing different temperature units.

    Length (Enum): Enum representing different units of length.
"""
from enum import Enum


class GenericUnit(Enum,):
    """
    GenericUnit is an enumeration class that represents various units of measurement
    used in the car connectivity context. This class is intended to be extended
    with specific units as needed.
    """

    def __str__(self) -> str:
        return self.value


class Temperature(GenericUnit,):
    """
    An enumeration representing different temperature units.

    Attributes:
        C (str): Celsius temperature unit represented by '°C'.
        F (str): Fahrenheit temperature unit represented by '°F'.
        K (str): Kelvin temperature unit represented by '°K'.
        INVALID (str): Represents an invalid temperature unit.
        UNKNOWN (str): Represents an unknown temperature unit.
    """
    C = '°C'
    F = '°F'
    K = '°K'
    INVALID = 'invalid'
    UNKNOWN = 'unknown temperature unit'


class Length(GenericUnit,):
    """
    Enum representing different units of length.

    Attributes:
        KM (str): Kilometers.
        MI (str): Miles.
        INVALID (str): Invalid length unit.
        UNKNOWN (str): Unknown length unit.
    """
    KM = 'km'
    MI = 'mi'
    INVALID = 'invalid'
    UNKNOWN = 'unknown length unit'


class Level(GenericUnit,):
    """
    A class representing a unit of measurement for levels.

    Attributes:
        PERCENTAGE (str): A string representing the percentage unit ('%').
        INVALID (str): A string representing an invalid unit.
        UNKNOWN (str): A string representing an unknown level unit.
    """
    PERCENTAGE = '%'
    INVALID = 'invalid'
    UNKNOWN = 'unknown level unit'


class Power(GenericUnit,):
    """
    A class representing a unit of measurement for power.
    """
    KW = 'KW'
    W = 'W'
    INVALID = 'invalid'
    UNKNOWN = 'unknown power unit'


class Energy(GenericUnit,):
    """
    A class representing a unit of measurement for energy.
    """
    KWH = 'KWH'
    WH = 'WH'
    INVALID = 'invalid'
    UNKNOWN = 'unknown energy unit'


class Current(GenericUnit,):
    """
    A class representing a unit of measurement for current.
    """
    A = 'A'
    INVALID = 'invalid'
    UNKNOWN = 'unknown current unit'


class Speed(GenericUnit,):
    """
    A class representing a unit of measurement for speed.
    """
    KMH = 'Km/h'
    MPH = 'Mph'
    INVALID = 'invalid'
    UNKNOWN = 'unknown speed unit'


class LatitudeLongitude(GenericUnit,):
    """
    A class representing a unit of measurement for levels.

    Attributes:
        PERCENTAGE (str): A string representing the percentage unit ('%').
        INVALID (str): A string representing an invalid unit.
        UNKNOWN (str): A string representing an unknown level unit.
    """
    DEGREE = '°'
    UNKNOWN = 'unknown level unit'



""" This module contains the base class for connector UI components. """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
import flask
import flask_login

from carconnectivity.util import config_remove_credentials

if TYPE_CHECKING:
    from typing import List, Dict, Union, Literal

    from carconnectivity_connectors.base.connector import BaseConnector


class BaseConnectorUI:
    """
    A base class for connector UI components.
    """
    def __init__(self, connector: BaseConnector, blueprint: flask.Blueprint):
        self.blueprint: flask.Blueprint = blueprint
        self.connector: BaseConnector = connector

        @self.blueprint.route('/log', methods=['GET'])
        @flask_login.login_required
        def log():
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            return flask.render_template('connectors/log.html', current_app=flask.current_app, connector=self.connector, formatter=formatter)

        @self.blueprint.route('/config', methods=['GET'])
        @flask_login.login_required
        def config():
            return flask.render_template('connectors/config.html', current_app=flask.current_app, connector=self.connector,
                                         config=config_remove_credentials(self.connector.active_config))

    def get_nav_items(self) -> List[Dict[Literal['text', 'url', 'sublinks', 'divider'], Union[str, List]]]:
        """
        Retrieve navigation items for the UI.

        This method should be implemented by subclasses to provide the navigation
        items required for the user interface. The navigation items are represented
        as a list of dictionaries, where each dictionary contains the following keys:
        - 'text': A string representing the display text of the navigation item.
        - 'url': A string representing the URL the navigation item points to.
        - 'sublinks': A list of sub-navigation items, each following the same structure.
        - 'divider': A string indicating a divider in the navigation menu.
        """
        return [{"text": "Config", "url": flask.url_for('connectors.'+self.blueprint.name+'.config')},
                {"text": "Log", "url": flask.url_for('connectors.'+self.blueprint.name+'.log')}]

    def get_title(self) -> str:
        """
        Retrieve the title for the connector UI.

        This method must be implemented by subclasses to provide a specific title.
        """
        raise NotImplementedError("Subclasses must provide get_title method")



"""Module for door classes."""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import EnumAttribute
from carconnectivity.commands import Commands

if TYPE_CHECKING:
    from typing import Optional, Dict
    from carconnectivity.vehicle import GenericVehicle


class Doors(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent all doors in the vehicle.
    """
    def __init__(self, vehicle: Optional[GenericVehicle] = None) -> None:
        if vehicle is None:
            raise ValueError('Cannot create doors without vehicle')
        super().__init__(object_id='doors', parent=vehicle)
        self.commands: Commands = Commands(parent=self)
        self.open_state = EnumAttribute("open_state", self, tags={'carconnectivity'})
        self.lock_state = EnumAttribute("lock_state", self, tags={'carconnectivity'})
        self.doors: Dict[str, Doors.Door] = {}

    # pylint: disable=duplicate-code
    class OpenState(Enum):
        """
        Enum for door open state.
        """
        OPEN = 'open'
        CLOSED = 'closed'
        UNSUPPORTED = 'unsupported'
        INVALID = 'invalid'
        UNKNOWN = 'unknown open state'
    # pylint: enable=duplicate-code

    class LockState(Enum):
        """
        Enum for door lock state.
        """
        LOCKED = 'locked'
        UNLOCKED = 'unlocked'
        INVALID = 'invalid'
        UNKNOWN = 'unknown lock state'

    class Door(GenericObject):
        """
        A class to represent a door in the vehicle.
        """
        def __init__(self, door_id: str, doors: Doors) -> None:
            super().__init__(object_id=door_id, parent=doors)
            self.door_id: str = door_id
            self.open_state = EnumAttribute("open_state", self, tags={'carconnectivity'})
            self.lock_state = EnumAttribute("lock_state", self, tags={'carconnectivity'})



"""Module for position classes."""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import EnumAttribute, FloatAttribute
from carconnectivity.units import LatitudeLongitude

if TYPE_CHECKING:
    from typing import Optional


class Position(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent a position.
    """
    def __init__(self, parent: Optional[GenericObject] = None) -> None:
        super().__init__(object_id='position', parent=parent)
        self.position_type: EnumAttribute = EnumAttribute("position_type", parent=self, value_type=Position.PositionType,
                                                          tags={'carconnectivity'})
        self.latitude: FloatAttribute = FloatAttribute("latitude", parent=self, minimum=-90, maximum=90, unit=LatitudeLongitude.DEGREE,
                                                       tags={'carconnectivity'})
        self.longitude: FloatAttribute = FloatAttribute("longitude", parent=self, minimum=-180, maximum=180, unit=LatitudeLongitude.DEGREE,
                                                        tags={'carconnectivity'})

    class PositionType(Enum):
        """
        Enum for position type.
        """
        PARKING = 'parking'
        DRIVING = 'driving'
        INVALID = 'invalid'
        UNKNOWN = 'unknown'



"""
This module defines classes related to batteries.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import TemperatureAttribute, EnergyAttribute
from carconnectivity.units import Temperature, Energy

if TYPE_CHECKING:
    from carconnectivity.drive import ElectricDrive


class Battery(GenericObject):
    """
    Represents the battery of a vehicle.
    """
    def __init__(self, drive: ElectricDrive) -> None:
        super().__init__(object_id='battery', parent=drive)
        self.total_capacity: EnergyAttribute = EnergyAttribute(name='total_capacity', parent=self, value=None, unit=Energy.KWH, minimum=0,
                                                               tags={'carconnectivity'})
        self.available_capacity: EnergyAttribute = EnergyAttribute(name='available_capacity', parent=self, value=None, unit=Energy.KWH,
                                                                   minimum=0, tags={'carconnectivity'})
        self.temperature: TemperatureAttribute = TemperatureAttribute(name="temperature", parent=self, value=None, unit=Temperature.C, tags={'carconnectivity'})
        self.temperature_min: TemperatureAttribute = TemperatureAttribute(name="temperature_min", parent=self, value=None, unit=Temperature.C,
                                                                          tags={'carconnectivity'})
        self.temperature_max: TemperatureAttribute = TemperatureAttribute(name="temperature_max", parent=self, value=None, unit=Temperature.C,
                                                                          tags={'carconnectivity'})



"""
Module for connector.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import EnumAttribute
if TYPE_CHECKING:
    from typing import Optional
    from carconnectivity.charging import Charging


class ChargingConnector(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent the charging connector of a vehicle.
    """
    def __init__(self, charging: Optional[Charging] = None) -> None:
        if charging is None:
            raise ValueError('Cannot create connector without charging')
        super().__init__(object_id='connector', parent=charging)
        self.delay_notifications = True
        self.connection_state: EnumAttribute = EnumAttribute("connection_state", parent=self, value_type=ChargingConnector.ChargingConnectorConnectionState,
                                                             tags={'carconnectivity'})
        self.lock_state: EnumAttribute = EnumAttribute("lock_state", parent=self, value_type=ChargingConnector.ChargingConnectorLockState,
                                                       tags={'carconnectivity'})
        self.external_power: EnumAttribute = EnumAttribute("external_power", parent=self, value_type=ChargingConnector.ExternalPower,
                                                           tags={'carconnectivity'})
        self.delay_notifications = False

    class ChargingConnectorConnectionState(Enum,):
        """
        Enum representing the connection state of the connector.

        Attributes:
            CONNECTED (str): The connector is connected.
            DISCONNECTED (str): The connector is disconnected.
            INVALID (str): The connector state is invalid.
            UNSUPPORTED (str): The connector state is unsupported.
            UNKNOWN (str): The connector state is unknown.
        """
        CONNECTED = 'connected'
        DISCONNECTED = 'disconnected'
        INVALID = 'invalid'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown unlock plug state'

    class ChargingConnectorLockState(Enum,):
        """
        Enum representing the lock state of a plug.

        Attributes:
            LOCKED (str): The plug is locked.
            UNLOCKED (str): The plug is unlocked.
            INVALID (str): The plug state is invalid.
            UNSUPPORTED (str): The plug state is unsupported.
            UNKNOWN (str): The plug state is unknown.
        """
        LOCKED = 'locked'
        UNLOCKED = 'unlocked'
        INVALID = 'invalid'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown unlock plug state'

    class ExternalPower(Enum,):
        """
        Enum representing the state of external power.

        Attributes:
            AVAILABLE (str): External power is available.
            ACTIVE (str): External power is used.
            UNAVAILABLE (str): External power is unavailable.
            INVALID (str): External power state is invalid.
            UNSUPPORTED (str): External power is unsupported.
            UNKNOWN (str): External power state is unknown.
        """
        AVAILABLE = 'available'
        ACTIVE = 'active'
        UNAVAILABLE = 'unavailable'
        INVALID = 'invalid'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown external power'



"""
Module for information about the vehicle maintenance
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import DateAttribute, RangeAttribute
from carconnectivity.units import Length


if TYPE_CHECKING:
    from carconnectivity.vehicle import GenericVehicle


class Maintenance(GenericObject):
    """
    Represents the maintenance of a vehicle.
    """
    def __init__(self, vehicle: GenericVehicle) -> None:
        super().__init__(object_id='maintenance', parent=vehicle)
        self.inspection_due_at = DateAttribute('inspection_due_at', parent=self, tags={'carconnectivity'})
        self.inspection_due_after = RangeAttribute('inspection_due_after', parent=self, tags={'carconnectivity'}, unit=Length.KM)
        self.oil_service_due_at = DateAttribute('oil_service_due_at', parent=self, tags={'carconnectivity'})
        self.oil_service_due_after = RangeAttribute('oil_service_due_after', parent=self, tags={'carconnectivity'}, unit=Length.KM)



"""Module for window heating classes."""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import EnumAttribute
from carconnectivity.commands import Commands

if TYPE_CHECKING:
    from typing import Optional, Dict
    from carconnectivity.vehicle import GenericVehicle


class WindowHeatings(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent all window heaters in the vehicle.
    """
    def __init__(self, vehicle: Optional[GenericVehicle] = None) -> None:
        if vehicle is None:
            raise ValueError('Cannot create window heatings without vehicle')
        super().__init__(object_id='window_heating', parent=vehicle)
        self.commands: Commands = Commands(parent=self)
        self.heating_state = EnumAttribute("heating_state", self, value_type=WindowHeatings.HeatingState, tags={'carconnectivity'})
        self.doors: Dict[str, WindowHeatings.WindowHeating] = {}

    class HeatingState(Enum):
        """
        Enum for heating state.
        """
        ON = 'on'
        OFF = 'off'
        UNSUPPORTED = 'unsupported'
        INVALID = 'invalid'
        UNKNOWN = 'unknown heating state'

    class WindowHeating(GenericObject):
        """
        A class to represent a window heating in the vehicle.
        """
        def __init__(self, window_id: str, window_heatings: WindowHeatings) -> None:
            super().__init__(object_id=window_id, parent=window_heatings)
            self.window_id: str = window_id
            self.heating_state = EnumAttribute("heating_state", self, value_type=WindowHeatings.HeatingState, tags={'carconnectivity'})



"""Module containing custom exceptions for the carconnectivity package."""
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


class CarConnectivityError(Exception):
    """
    Base exception class for the carconnectivity package.
    """


class ConfigurationError(CarConnectivityError):
    """
    Exception raised for problems with the configuration.
    """


class RetrievalError(CarConnectivityError):
    """
    Exception raised for errors that occur during data retrieval.
    """


class TooManyRequestsError(RetrievalError):
    """
    Exception raised when too many requests are made to a service.

    This error is typically used to indicate that the client should
    slow down or wait before making further requests.
    """
    def __init__(self, *args: object, retry_after: Optional[int] = None) -> None:
        super().__init__(*args)
        self.retry_after: Optional[int] = retry_after


class SetterError(CarConnectivityError):
    """
    Exception raised for errors in the setter methods.
    """


class CommandError(SetterError):
    """
    Exception raised for errors when controlling something.
    """


class AuthenticationError(CarConnectivityError):
    """
    Exception raised for errors in the authentication process.
    """


class TemporaryAuthenticationError(AuthenticationError):
    """
    Exception raised for temporary authentication errors.

    This exception is a subclass of `AuthenticationError` and is used to indicate
    that an authentication error occurred, but it is temporary and may be resolved
    by retrying the operation.
    """


class APICompatibilityError(CarConnectivityError):
    """
    Exception raised for errors in the API compatibility.

    This exception is used to indicate that there is an issue with the compatibility
    of the API, such as when an expected API version is not supported or when there
    are breaking changes in the API.
    """


class APIError(CarConnectivityError):
    """
    Exception raised for errors that occur during API calls.

    This is a generic exception that can be used to indicate any kind of error
    related to API operations.
    """



"""This module defines the classes that represent attributes in the CarConnectivity system."""
from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Union

from enum import Enum
import argparse
import logging

from carconnectivity.commands import GenericCommand
from carconnectivity.objects import GenericObject
from carconnectivity.errors import SetterError
from carconnectivity.units import Temperature
from carconnectivity.climatization import Climatization
from carconnectivity.util import ThrowingArgumentParser

if TYPE_CHECKING:
    from carconnectivity.objects import Optional

LOG: logging.Logger = logging.getLogger("carconnectivity")


class ClimatizationStartStopCommand(GenericCommand):
    """
    ClimatizationStartStopCommand is a command class for starting or stopping the climatization system.

    Command (Enum): Enum class representing different commands for climatization.

    """
    def __init__(self, name: str = 'start-stop', parent: Optional[GenericObject] = None) -> None:
        super().__init__(name=name, parent=parent)

    @property
    def value(self) -> Optional[Union[str, Dict]]:
        return super().value

    # pylint: disable=duplicate-code
    @value.setter
    def value(self, new_value: Optional[Union[str, Dict]]) -> None:
        # Execute early hooks before parsing the value
        new_value = self._execute_on_set_hook(new_value, early_hook=True)
        if isinstance(new_value, str):
            parser = ThrowingArgumentParser(prog='', add_help=False, exit_on_error=False)
            parser.add_argument('command', help='Command to execute', type=ClimatizationStartStopCommand.Command,
                                choices=list(ClimatizationStartStopCommand.Command))
            if self.parent is not None and isinstance((climatization := self.parent.parent), Climatization) and climatization.settings is not None \
                    and (target_temperature_attribute := climatization.settings.target_temperature) is not None \
                    and target_temperature_attribute.value is not None:
                default_temperature = target_temperature_attribute.value
                default_temperature_unit = target_temperature_attribute.unit
            else:
                default_temperature = 25
                default_temperature_unit = Temperature.C
            parser.add_argument('--target-temperature', dest='target_temperature', help='Target temperature for climatization', type=float, required=False,
                                default=default_temperature)
            parser.add_argument('--target-temperature-unit', dest='target_temperature_unit', help='Target temperature for climatization', type=Temperature,
                                required=False, choices=list(Temperature), default=default_temperature_unit)
            try:
                args = parser.parse_args(new_value.split(sep=' '))
            except argparse.ArgumentError as e:
                raise SetterError(f'Invalid format for ClimatizationStartStopCommand: {e.message} {parser.format_usage()}') from e

            newvalue_dict = {}
            newvalue_dict['command'] = args.command
            newvalue_dict['target_temperature'] = args.target_temperature
            newvalue_dict['target_temperature_unit'] = args.target_temperature_unit
            new_value = newvalue_dict
        elif isinstance(new_value, dict):
            if 'command' in new_value and isinstance(new_value['command'], str):
                if new_value['command'] in ClimatizationStartStopCommand.Command:
                    new_value['command'] = ClimatizationStartStopCommand.Command(new_value['command'])
                else:
                    raise ValueError('Invalid value for ClimatizationStartStopCommand. '
                                     f'Command must be one of {ClimatizationStartStopCommand.Command}')
        if self._is_changeable:
            # Execute late hooks before setting the value
            new_value = self._execute_on_set_hook(new_value, early_hook=False)
            self._set_value(new_value)
        else:
            raise TypeError('You cannot use this command. Command is not implemented.')
    # pylint: enable=duplicate-code

    class Command(Enum):
        """
        Enum class representing different commands for climatization.

        Attributes:
            START (str): Command to start the climatization.
            STOP (str): Command to stop the climatization.
        """
        START = 'start'
        STOP = 'stop'

        def __str__(self) -> str:
            return self.value


class ChargingStartStopCommand(GenericCommand):
    """
    ChargingStartStopCommand is a command class for starting or stopping the charging.

    Command (Enum): Enum class representing different commands for charging.

    """
    def __init__(self, name: str = 'start-stop', parent: Optional[GenericObject] = None) -> None:
        super().__init__(name=name, parent=parent)

    @property
    def value(self) -> Optional[Union[str, Dict]]:
        return super().value

    @value.setter
    def value(self, new_value: Optional[Union[str, Dict]]) -> None:
        # Execute early hooks before parsing the value
        new_value = self._execute_on_set_hook(new_value, early_hook=True)
        if isinstance(new_value, str):
            parser = ThrowingArgumentParser(prog='', add_help=False, exit_on_error=False)
            parser.add_argument('command', help='Command to execute', type=ChargingStartStopCommand.Command,
                                choices=list(ChargingStartStopCommand.Command))
            try:
                args = parser.parse_args(new_value.split(sep=' '))
            except argparse.ArgumentError as e:
                raise SetterError(f'Invalid format for ChargingStartStopCommand: {e.message} {parser.format_usage()}') from e

            newvalue_dict = {}
            newvalue_dict['command'] = args.command
            new_value = newvalue_dict
        elif isinstance(new_value, dict):
            if 'command' in new_value and isinstance(new_value['command'], str):
                if new_value['command'] in ChargingStartStopCommand.Command:
                    new_value['command'] = ChargingStartStopCommand.Command(new_value['command'])
                else:
                    raise ValueError('Invalid value for ChargingStartStopCommand. '
                                     f'Command must be one of {ChargingStartStopCommand.Command}')
        if self._is_changeable:
            # Execute late hooks before setting the value
            new_value = self._execute_on_set_hook(new_value, early_hook=False)
            self._set_value(new_value)
        else:
            raise TypeError('You cannot set this attribute. Attribute is not mutable.')

    class Command(Enum):
        """
        Enum class representing different commands for charging.

        Attributes:
            START (str): Command to start the charging.
            STOP (str): Command to stop the charging.
        """
        START = 'start'
        STOP = 'stop'

        def __str__(self) -> str:
            return self.value


class HonkAndFlashCommand(GenericCommand):
    """
    HonkAndFlashCommand is a command class for honking and flashing the lights.
    """
    def __init__(self, name: str = 'honk-flash', parent: Optional[GenericObject] = None, with_duration: bool = False) -> None:
        super().__init__(name=name, parent=parent)
        self.with_duration: bool = with_duration

    @property
    def value(self) -> Optional[Union[str, Dict]]:
        return super().value

    @value.setter
    def value(self, new_value: Optional[Union[str, Dict]]) -> None:
        # Execute early hooks before parsing the value
        new_value = self._execute_on_set_hook(new_value, early_hook=True)
        if isinstance(new_value, str):
            parser = ThrowingArgumentParser(prog='', add_help=False, exit_on_error=False)
            parser.add_argument('command', help='Command to execute', type=HonkAndFlashCommand.Command,
                                choices=list(HonkAndFlashCommand.Command))
            if self.with_duration:
                parser.add_argument('--duration', dest='duration', help='Duration for honking and flashing in seconds', type=int, required=False)
            try:
                args = parser.parse_args(new_value.split(sep=' '))
            except argparse.ArgumentError as e:
                raise SetterError(f'Invalid format for HonkAndFlashCommand: {e.message} {parser.format_usage()}') from e

            newvalue_dict = {}
            newvalue_dict['command'] = args.command
            if self.with_duration:
                newvalue_dict['duration'] = args.duration
            new_value = newvalue_dict
        elif isinstance(new_value, dict):
            if 'command' in new_value and isinstance(new_value['command'], str):
                if new_value['command'] in HonkAndFlashCommand.Command:
                    new_value['command'] = HonkAndFlashCommand.Command(new_value['command'])
                else:
                    raise ValueError('Invalid value for HonkAndFlashCommand. '
                                     f'Command must be one of {HonkAndFlashCommand.Command}')
        if self._is_changeable:
            # Execute late hooks before setting the value
            new_value = self._execute_on_set_hook(new_value, early_hook=False)
            self._set_value(new_value)
        else:
            raise TypeError('You cannot set this attribute. Attribute is not mutable.')

    class Command(Enum):
        """
        Enum class representing different commands for honking and flashing the lights.

        Attributes:
            FLASH (str): Command to flash the lights.
            HONK_AND_FLASH (str): Command to honk and flash the lights.
        """
        FLASH = 'flash'
        HONK_AND_FLASH = 'honk-and-flash'

        def __str__(self) -> str:
            return self.value


class LockUnlockCommand(GenericCommand):
    """
    LockUnlockCommand is a command class for locking and unlocking the vehicle.

    Command (Enum): Enum class representing different commands for locking.

    """
    def __init__(self, name: str = 'lock-unlock', parent: Optional[GenericObject] = None) -> None:
        super().__init__(name=name, parent=parent)

    @property
    def value(self) -> Optional[Union[str, Dict]]:
        return super().value

    @value.setter
    def value(self, new_value: Optional[Union[str, Dict]]) -> None:
        # Execute early hooks before parsing the value
        new_value = self._execute_on_set_hook(new_value, early_hook=True)
        if isinstance(new_value, str):
            parser = ThrowingArgumentParser(prog='', add_help=False, exit_on_error=False)
            parser.add_argument('command', help='Command to execute', type=LockUnlockCommand.Command,
                                choices=list(LockUnlockCommand.Command))
            try:
                args = parser.parse_args(new_value.split(sep=' '))
            except argparse.ArgumentError as e:
                raise SetterError(f'Invalid format for LockUnlockCommand: {e.message} {parser.format_usage()}') from e

            newvalue_dict = {}
            newvalue_dict['command'] = args.command
            new_value = newvalue_dict
        elif isinstance(new_value, dict):
            if 'command' in new_value and isinstance(new_value['command'], str):
                if new_value['command'] in LockUnlockCommand.Command:
                    new_value['command'] = LockUnlockCommand.Command(new_value['command'])
                else:
                    raise ValueError('Invalid value for LockUnlockCommand. '
                                     f'Command must be one of {LockUnlockCommand.Command}')
        if self._is_changeable:
            # Execute late hooks before setting the value
            new_value = self._execute_on_set_hook(new_value, early_hook=False)
            self._set_value(new_value)
        else:
            raise TypeError('You cannot set this attribute. Attribute is not mutable.')

    class Command(Enum):
        """
        Enum class representing different commands for locking.

        Attributes:
            LOCK (str): Command to lock the vehicle.
            UNLOCK (str): Command to unlocking the vehicle.
        """
        LOCK = 'lock'
        UNLOCK = 'unlock'

        def __str__(self) -> str:
            return self.value


class WakeSleepCommand(GenericCommand):
    """
    WakeCommand is a command class for waking up a vehicle.

    Command (Enum): Enum class representing different commands for wake/sleep.

    """
    def __init__(self, name: str = 'wake-sleep', parent: Optional[GenericObject] = None) -> None:
        super().__init__(name=name, parent=parent)

    @property
    def value(self) -> Optional[Union[str, Dict]]:
        return super().value

    # pylint: disable=duplicate-code
    @value.setter
    def value(self, new_value: Optional[Union[str, Dict]]) -> None:
        # Execute early hooks before parsing the value
        new_value = self._execute_on_set_hook(new_value, early_hook=True)
        if isinstance(new_value, str):
            parser = ThrowingArgumentParser(prog='', add_help=False, exit_on_error=False)
            parser.add_argument('command', help='Command to execute', type=WakeSleepCommand.Command,
                                choices=list(WakeSleepCommand.Command))
            try:
                args = parser.parse_args(new_value.split(sep=' '))
            except argparse.ArgumentError as e:
                raise SetterError(f'Invalid format for WakeSleepCommand: {e.message} {parser.format_usage()}') from e

            newvalue_dict = {}
            newvalue_dict['command'] = args.command
            new_value = newvalue_dict
        elif isinstance(new_value, dict):
            if 'command' in new_value and isinstance(new_value['command'], str):
                if new_value['command'] in WakeSleepCommand.Command:
                    new_value['command'] = WakeSleepCommand.Command(new_value['command'])
                else:
                    raise ValueError('Invalid value for WakeSleepCommand. '
                                     f'Command must be one of {WakeSleepCommand.Command}')
        if self._is_changeable:
            # Execute late hooks before setting the value
            new_value = self._execute_on_set_hook(new_value, early_hook=False)
            self._set_value(new_value)
        else:
            raise TypeError('You cannot use this command. Command is not implemented.')
    # pylint: enable=duplicate-code

    class Command(Enum):
        """
        Enum representing different commands for car connectivity.

        Attributes:
            WAKE (str): Command to wake up the vehicle.
            SLEEP (str): Command to put the vehicle to sleep.
        """
        WAKE = 'wake'
        SLEEP = 'sleep'

        def __str__(self) -> str:
            return self.value



"""Contains the main class to interact with the carconnectivity library."""
from __future__ import annotations
from typing import TYPE_CHECKING

import importlib
import pkgutil

import re
import json
import logging
import os
from datetime import datetime, timezone
from cryptography.fernet import Fernet

import carconnectivity_connectors
import carconnectivity_plugins

from carconnectivity.objects import GenericObject
from carconnectivity.garage import Garage
from carconnectivity.json_util import ExtendedEncoder
from carconnectivity.errors import ConfigurationError
from carconnectivity.connectors import Connectors
from carconnectivity.plugins import Plugins
from carconnectivity.attributes import StringAttribute
from carconnectivity._version import __version__
from carconnectivity.util import LogMemoryHandler

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterator

    from types import ModuleType

    from carconnectivity_connectors.base.connector import BaseConnector
    from carconnectivity_plugins.base.plugin import BasePlugin

LOG: logging.Logger = logging.getLogger("carconnectivity")

TOKENSTORE_FORMAT_VERSION: str = '1.0'
CACHE_FORMAT_VERSION: str = '1.0'
TOKENSTORE_KEY: str = '5weee2AYwL08LfVsDzzzDL82ffN6lWgwjUHYPzdzZBk='
CACHE_KEY: str = '5weee2AYwL08LfVsDzzzDL82ffN6lWgwjUHYPzdzZBk='


def __iter_namespace(ns_pkg) -> Iterator[pkgutil.ModuleInfo]:
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


discovered_connectors: Dict[str, ModuleType] = {
    name: importlib.import_module('.connector', name)
    for finder, name, ispkg
    in __iter_namespace(carconnectivity_connectors)
}

discovered_plugins: Dict[str, ModuleType] = {
    name: importlib.import_module('.plugin', name)
    for finder, name, ispkg
    in __iter_namespace(carconnectivity_plugins)
}


class CarConnectivity(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    CarConnectivity class is the main class to interact with the carconnectivity library.

    Attributes:
        config (Dict): Configuration dictionary for car connectivity.
        connectors (List[BaseConnector]): List of connector instances.
        plugins (List[BasePlugin]): List of plugin instances.
        garage (Garage): Instance of the Garage class.
    """
    # pylint: disable=too-many-statements, too-many-branches, too-many-locals
    def __init__(self, config: Dict, tokenstore_file: Optional[str] = None, cache_file: Optional[str] = None) -> None:
        """
        Initialize the CarConnectivity object.

        Args:
            config (Dict): Configuration dictionary for car connectivity.
            tokenstore_file (Optional[str]): Path to the token store file. Defaults to None.

        Raises:
            ValueError: If the configuration is invalid.
        """
        super().__init__(object_id='', parent=None)
        self.delay_notifications = True
        self.__cache: Dict[str, Any] = {}
        self.__tokenstore: Dict[str, Any] = {}
        self.__tokenstore_file: Optional[str] = tokenstore_file
        self.__cache_file: Optional[str] = cache_file

        self.config: Dict[Any, Any] = config
        self.connectors: Connectors = Connectors(car_connectivity=self)
        self.plugins: Plugins = Plugins(car_connectivity=self)
        self.garage: Garage = Garage(self)
        self.log_storage: LogMemoryHandler = LogMemoryHandler()

        self.version: StringAttribute = StringAttribute(name="version", parent=self, value=__version__, tags={'carconnectivity'})

        if 'carConnectivity' not in config:
            raise ConfigurationError("Invalid configuration: 'carConnectivity' is missing")
        # Configure logging
        if 'log_level' in config['carConnectivity'] and config['carConnectivity']['log_level'] is not None:
            config['carConnectivity']['log_level'] = config['carConnectivity']['log_level'].upper()
            if config['carConnectivity']['log_level'] in logging._nameToLevel:
                LOG.setLevel(config['carConnectivity']['log_level'])
            else:
                raise ConfigurationError(f'Invalid log level: "{config["carConnectivity"]["log_level"]}" not in {list(logging._nameToLevel.keys())}')
        if 'log_format' in config['carConnectivity'] and config['carConnectivity']['log_format'] is not None:
            log_format: str = config['carConnectivity']['log_format']
        else:
            log_format: str = '%(asctime)s:%(name)s:%(levelname)s:%(module)s:%(message)s'
        formatter = logging.Formatter(log_format)
        if 'log_date_format' in config['carConnectivity'] and config['carConnectivity']['log_date_format'] is not None:
            formatter.datefmt = config['carConnectivity']['log_date_format']
        else:
            formatter.datefmt = '%Y-%m-%dT%H:%M:%S%z'
        for handler in logging.getLogger().handlers:
            handler.setFormatter(formatter)
        LOG.addHandler(self.log_storage)
        self.log_storage.setFormatter(formatter)

        # pylint: disable=too-few-public-methods
        class NoPluginsConnectorsAPIDebug(logging.Filter):
            """
            A logging filter that excludes connector and plugin messages from the logs.

            Methods:
                filter(record): Determines if the log record should be logged.
            """
            def filter(self, record):
                pattern = re.compile(r'carconnectivity\.(connectors|plugins)\..*-api-debug')
                return not pattern.match(record.name)

        #  Disable logging for plugins and connectors
        self.log_storage.addFilter(NoPluginsConnectorsAPIDebug())

        if self.__tokenstore_file is not None:
            try:
                with open(file=self.__tokenstore_file, mode='r', encoding='utf8') as file:
                    tokenstore_file_dict: Dict[str, Any] = json.load(file)
                    if 'format_version' not in tokenstore_file_dict or tokenstore_file_dict['format_version'] != TOKENSTORE_FORMAT_VERSION:
                        LOG.info('Tokenstore file has wrong format version, ignoring it')
                        self.__tokenstore = {}
                    else:
                        if 'tokenstore_encrypted' in self.config['carConnectivity'] and not self.config['carConnectivity']['tokenstore_encrypted']:
                            self.__tokenstore = tokenstore_file_dict['tokenstore']
                        else:
                            fernet = Fernet(TOKENSTORE_KEY.encode('utf-8'))
                            self.__tokenstore = json.loads(fernet.decrypt(tokenstore_file_dict['tokenstore'].encode('utf-8')).decode('utf-8'))
            except json.JSONDecodeError as err:
                LOG.info('Could not use tokenstore from file %s (%s)', tokenstore_file, err.msg)
                self.__tokenstore = {}
            except FileNotFoundError as err:
                LOG.info('Could not use tokenstore from file %s (%s)', tokenstore_file, err)
                self.__tokenstore = {}
        else:
            self.__tokenstore = {}

        # Fill Cache
        if self.__cache_file is not None:
            LOG.info('Reading cachefile %s', cache_file)
            try:
                with open(self.__cache_file, 'r', encoding='utf8') as file:
                    cache_file_dict: Dict[str, Any] = json.load(file)
                    if 'format_version' not in cache_file_dict or cache_file_dict['format_version'] != CACHE_FORMAT_VERSION:
                        LOG.info('Cache file has wrong format version, ignoring it')
                        self.__cache = {}
                    else:
                        if 'cache_encrypted' in self.config['carConnectivity'] and not self.config['carConnectivity']['cache_encrypted']:
                            self.__cache = cache_file_dict['cache']
                        else:
                            fernet = Fernet(CACHE_KEY.encode('utf-8'))
                            self.__cache = json.loads(fernet.decrypt(cache_file_dict['cache'].encode('utf-8')).decode('utf-8'))
            except json.decoder.JSONDecodeError:
                LOG.error('Cachefile %s seems corrupted will delete it and try to create a new one. '
                          'If this problem persists please check if a problem with your disk exists.', self.__cache_file)
                os.remove(self.__cache_file)
                self.__cache = {}
            except FileNotFoundError:
                self.__cache = {}

        if 'connectors' in config['carConnectivity']:
            for connector_config in config['carConnectivity']['connectors']:
                if 'type' not in connector_config:
                    raise ConfigurationError("Invalid configuration: 'type' is missing in connector")
                if f"carconnectivity_connectors.{connector_config['type']}" not in discovered_connectors:
                    raise ConfigurationError(f"Invalid configuration: connector type '{connector_config['type']}' is not known")
                if 'disabled' in connector_config and connector_config['disabled']:
                    LOG.info('Skipping disabled connector %s', connector_config['type'])
                    continue
                connector_class = getattr(discovered_connectors['carconnectivity_connectors.' + connector_config['type']], 'Connector')
                if 'connector_id' in connector_config and connector_config['connector_id'] is not None:
                    connector_id = connector_config['connector_id']
                else:
                    connector_id = connector_config['type']
                if connector_id in self.connectors.connectors:
                    raise ConfigurationError(f"Invalid configuration: connector '{connector_id}' is not unique, set a 'connector_id' in configuration")
                connector: BaseConnector = connector_class(connector_id=connector_id, car_connectivity=self, config=connector_config['config'])
                self.connectors.connectors[connector_id] = connector
        if 'plugins' in config['carConnectivity']:
            for plugin_config in config['carConnectivity']['plugins']:
                if 'type' not in plugin_config:
                    raise ConfigurationError("Invalid configuration: 'type' is missing in plugin")
                if f"carconnectivity_plugins.{plugin_config['type']}" not in discovered_plugins:
                    raise ConfigurationError(f"Invalid configuration: plugin type '{plugin_config['type']}' is not known")
                if 'disabled' in plugin_config and plugin_config['disabled']:
                    LOG.info('Skipping disabled plugin %s', plugin_config['type'])
                    continue
                plugin_class = getattr(discovered_plugins['carconnectivity_plugins.' + plugin_config['type']], 'Plugin')
                if 'plugin_id' in plugin_config and plugin_config['plugin_id'] is not None:
                    plugin_id: str = plugin_config['plugin_id']
                else:
                    plugin_id: str = plugin_config['type']
                if plugin_id in self.plugins.plugins:
                    raise ConfigurationError(f"Invalid configuration: connector '{plugin_id}' is not unique, set a 'connector_id' in configuration")
                plugin: BasePlugin = plugin_class(plugin_id=plugin_id, car_connectivity=self, config=plugin_config['config'])
                self.plugins.plugins[plugin_id] = plugin
        self.delay_notifications = False

    def fetch_all(self) -> None:
        """
        Fetch data from all connectors.

        This method iterates over all connectors in the `self.connectors` list
        and calls their `fetch_all` method to retrieve data.
        """
        for connector in self.connectors.connectors.values():
            connector.fetch_all()

    def persist(self) -> None:
        """
        Persist the token store to a file.

        This method writes the token store to a specified file in JSON format.
        If the token store or the file path is not set, the method does nothing.
        If an error occurs during writing, it logs an error message.
        """
        if self.__tokenstore and self.__tokenstore_file:
            try:
                with open(file=self.__tokenstore_file, mode='w', encoding='utf8') as file:
                    tokenstore_file_dict: Dict[str, Any] = {}
                    tokenstore_file_dict['format_version'] = TOKENSTORE_FORMAT_VERSION
                    tokenstore_file_dict['date'] = datetime.now(tz=timezone.utc).isoformat()
                    if 'tokenstore_encrypted' in self.config['carConnectivity'] and not self.config['carConnectivity']['tokenstore_encrypted']:
                        tokenstore_file_dict['tokenstore'] = self.__tokenstore
                    else:
                        fernet = Fernet(TOKENSTORE_KEY.encode('utf-8'))
                        tokenstore_file_dict['tokenstore'] = fernet.encrypt(json.dumps(self.__tokenstore, cls=ExtendedEncoder).encode('utf-8')).decode('utf-8')
                    json.dump(tokenstore_file_dict, file)
                LOG.info('Writing tokenstore to file %s', self.__tokenstore_file)
            except ValueError as err:  # pragma: no cover
                LOG.info('Could not write tokenstore to file %s (%s)', self.__tokenstore_file, err)

        # Persist cache
        if self.__cache and self.__cache_file:
            LOG.info('Writing cachefile %s', self.__cache_file)
            with open(file=self.__cache_file, mode='w', encoding='utf8') as file:
                cache_file_dict: Dict[str, Any] = {}
                cache_file_dict['format_version'] = CACHE_FORMAT_VERSION
                cache_file_dict['date'] = datetime.now(tz=timezone.utc).isoformat()
                if 'cache_encrypted' in self.config['carConnectivity'] and not self.config['carConnectivity']['cache_encrypted']:
                    cache_file_dict['cache'] = self.__cache
                else:
                    fernet = Fernet(CACHE_KEY.encode('utf-8'))
                    cache_file_dict['cache'] = fernet.encrypt(json.dumps(self.__cache).encode('utf-8')).decode('utf-8')
                json.dump(cache_file_dict, file, cls=ExtendedEncoder)

    def startup(self) -> None:
        """
        Start all connectors and plugins.

        This method iterates over all connectors in the `self.connectors` list and
        calls their `startup` method.
        """
        for connector in self.connectors.connectors.values():
            connector.startup()
        for plugin in self.plugins.plugins.values():
            plugin.startup()

    def shutdown(self) -> None:
        """
        Shuts down all connectors and persists the current state.

        This method iterates over all connectors in the `self.connectors` list and
        calls their `shutdown` method. After all connectors have been shut down,
        it calls the `persist` method to save the current state.
        """
        for connector in self.connectors.connectors.values():
            connector.shutdown()
        for plugin in self.plugins.plugins.values():
            plugin.shutdown()
        self.persist()

    def get_tokenstore(self) -> Dict[str, Any]:
        """
        Retrieve the token store.

        Returns:
            Dict[str, Any]: A dictionary containing the token store.
        """
        return self.__tokenstore

    def get_cache(self) -> Dict[str, Any]:
        """
        Retrieve the cache.

        Returns:
            Dict[str, Any]: The current cache stored in the object.
        """
        return self.__cache

    def get_garage(self) -> Optional[Garage]:
        """
        Retrieve the garage associated with the car connectivity.

        Returns:
            Garage | None: The garage object if available, otherwise None.
        """
        return self.garage

    def is_healthy(self) -> bool:
        """
        Returns whether the carconnectivity instance and its connectors and plugins is healthy.

        Returns:
            bool: True if carconnectivity is healthy, False otherwise.
        """
        for connector in self.connectors.connectors.values():
            if not connector.is_healthy():
                return False
        for plugin in self.plugins.plugins.values():
            if not plugin.is_healthy():
                return False
        return True



"""Module containing the BasePlugin class that needs to be extended to implement a new plugin."""
# pylint: disable=duplicate-code
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import StringAttribute, BooleanAttribute
from carconnectivity.errors import ConfigurationError
from carconnectivity.util import LogMemoryHandler

if TYPE_CHECKING:
    from typing import Dict, Any

    from carconnectivity.carconnectivity import CarConnectivity


class BasePlugin(GenericObject):  # pylint: disable=too-few-public-methods
    """BaseConnector is a base class for plugins in the CarConnectivity system.

    Attributes:

    Methods:
        __init__(car_connectivity: CarConnectivity, config: Dict) -> None:
        shutdown() -> None:
            Placeholder method for shutting down the plugin.
    """
    def __init__(self, plugin_id: str, car_connectivity: CarConnectivity, config: Dict, log: logging.Logger) -> None:
        """
        Initializes the connector with the given CarConnectivity instance and configuration.

        Args:
            car_connectivity (CarConnectivity): The instance in which the plugin is running.
            config (Dict): A dictionary containing the configuration parameters for this plugin only.
        """
        super().__init__(object_id=plugin_id, parent=car_connectivity.plugins)
        self.car_connectivity: CarConnectivity = car_connectivity
        self.active_config: Dict[str, Any] = {}
        self.log_storage: LogMemoryHandler = LogMemoryHandler()
        self.log_level = StringAttribute(name="log_level", parent=self, tags={'carconnectivity'})
        self.version = StringAttribute(name="version", parent=self, value=self.get_version(), tags={'carconnectivity'})
        self.healthy: BooleanAttribute = BooleanAttribute(name="healthy", parent=self, tags={'carconnectivity'})

        # Configure logging
        if 'log_level' in config and config['log_level'] is not None:
            self.active_config['log_level'] = config['log_level'].upper()
            if self.active_config['log_level'] in logging._nameToLevel:
                log.setLevel(self.active_config['log_level'])
                self.log_level._set_value(self.active_config['log_level'])  # pylint: disable=protected-access
            else:
                raise ConfigurationError(f'Invalid log level: "{self.active_config["log_level"]}" not in {list(logging._nameToLevel.keys())}')
        log.addHandler(self.log_storage)

    def startup(self) -> None:
        """
        Starts up the plugin.

        This method should be overridden by subclasses to implement any necessary
        startup procedures for the plugin. If threads are needed they should be started here.
        """

    def shutdown(self) -> None:
        """
        Shuts down the plugin.

        This method should be overridden by subclasses to implement any necessary
        cleanup or shutdown procedures for the plugin. If threads were started in startup() they should be stopped here.
        If data needs to be persisted, it should be done here.
        """

    def get_version(self) -> str:
        """
        Returns the version of the plugin.

        Returns:
            str: The version of the plugin.
        """
        raise NotImplementedError("Method get_version() must be implemented by plugin")

    def get_type(self) -> str:
        """
        Returns the type of the plugin.

        Returns:
            str: The type of the plugin.
        """
        raise NotImplementedError("Method get_type() must be implemented by plugin")

    def is_healthy(self) -> bool:
        """
        Returns whether the plugin is healthy.

        Returns:
            bool: True if the plugin is healthy, False otherwise.
        """
        if self.healthy.enabled and self.healthy.value is not None:
            return self.healthy.value
        return False

    def get_name(self) -> str:
        """
        Returns the user readable name of the plugin.
        If not implemented by the plugin, fallback is the ID of the plugin.

        Returns:
            str: The name of the plugin.
        """
        return self.id



""" Base class for plugin user interfaces. """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
import flask
import flask_login

from carconnectivity.util import config_remove_credentials

if TYPE_CHECKING:
    from typing import List, Dict, Union, Literal

    from carconnectivity_plugins.base.plugin import BasePlugin


class BasePluginUI:
    """
    Base class for plugin user interfaces.

    This class provides a blueprint for creating user interfaces for plugins. It defines
    the basic structure and methods that must be implemented by subclasses to provide
    specific navigation items and titles for the user interface.
    """
    def __init__(self, plugin: BasePlugin, blueprint: flask.Blueprint):
        self.blueprint: flask.Blueprint = blueprint
        self.plugin: BasePlugin = plugin

        @self.blueprint.route('/log', methods=['GET'])
        @flask_login.login_required
        def log():
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            return flask.render_template('plugins/log.html', current_app=flask.current_app, plugin=self.plugin, formatter=formatter)

        @self.blueprint.route('/config', methods=['GET'])
        @flask_login.login_required
        def config():
            return flask.render_template('plugins/config.html', current_app=flask.current_app, plugin=self.plugin,
                                         config=config_remove_credentials(self.plugin.active_config))

    def get_nav_items(self) -> List[Dict[Literal['text', 'url', 'sublinks', 'divider'], Union[str, List]]]:
        """
        Retrieve navigation items for the UI.

        This method should be implemented by subclasses to provide the navigation
        items required for the user interface. The navigation items are represented
        as a list of dictionaries, where each dictionary contains the following keys:
        - 'text': A string representing the display text of the navigation item.
        - 'url': A string representing the URL the navigation item points to.
        - 'sublinks': A list of sub-navigation items, each following the same structure.
        - 'divider': A string indicating a divider in the navigation menu.
        """
        return [{"text": "Config", "url": flask.url_for('plugins.'+self.blueprint.name+'.config')},
                {"text": "Log", "url": flask.url_for('plugins.'+self.blueprint.name+'.log')}]

    def get_title(self) -> str:
        """
        Retrieve the title for the connector UI.

        This method must be implemented by subclasses to provide a specific title.
        """
        raise NotImplementedError("Subclasses must provide get_title method")



"""Module containing the BaseConnector class that needs to be extended to implement a new connector."""
# pylint: disable=duplicate-code
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import StringAttribute, DateAttribute, BooleanAttribute
from carconnectivity.errors import ConfigurationError
from carconnectivity.util import LogMemoryHandler

if TYPE_CHECKING:
    from typing import Dict, Any, Optional

    from carconnectivity.carconnectivity import CarConnectivity


class BaseConnector(GenericObject):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """BaseConnector is a base class for connectors in the CarConnectivity system.

    Attributes:

    Methods:
        __init__(car_connectivity: CarConnectivity, config: Dict) -> None:
        shutdown() -> None:
            Placeholder method for shutting down the connector.
    """
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, connector_id: str, car_connectivity: CarConnectivity, config: Dict, log: logging.Logger, api_log: Optional[logging.Logger]) -> None:
        """
        Initializes the connector with the given CarConnectivity instance and configuration.

        Args:
            car_connectivity (CarConnectivity): The instance in which the connector is running.
            config (Dict): A dictionary containing the configuration parameters for this connector only.
        """
        super().__init__(object_id=connector_id, parent=car_connectivity.connectors)
        self.car_connectivity: CarConnectivity = car_connectivity
        self.active_config: Dict[str, Any] = {}
        self.log_storage: LogMemoryHandler = LogMemoryHandler()
        self.api_log_storage: LogMemoryHandler = LogMemoryHandler()
        self.log_level = StringAttribute(name="log_level", parent=self, tags={'carconnectivity'})
        self.version = StringAttribute(name="version", parent=self, value=self.get_version(), tags={'carconnectivity'})
        self.last_update: DateAttribute = DateAttribute(name="last_update", parent=self, tags={'carconnectivity'})
        self.healthy: BooleanAttribute = BooleanAttribute(name="healthy", parent=self, tags={'carconnectivity'})

        # Configure logging
        if 'log_level' in config and config['log_level'] is not None:
            self.active_config['log_level'] = config['log_level'].upper()
            if self.active_config['log_level'] in logging._nameToLevel:
                log.setLevel(self.active_config['log_level'])
                self.log_level._set_value(self.active_config['log_level'])  # pylint: disable=protected-access
                logging.getLogger('requests').setLevel(self.active_config['log_level'])
                logging.getLogger('urllib3').setLevel(self.active_config['log_level'])
                logging.getLogger('oauthlib').setLevel(self.active_config['log_level'])
            else:
                raise ConfigurationError(f'Invalid log level: "{self.active_config["log_level"]}" not in {list(logging._nameToLevel.keys())}')
        if api_log is not None and 'api_log_level' in config and config['api_log_level'] is not None:
            self.active_config['api_log_level'] = config['api_log_level'].upper()
            if self.active_config['api_log_level'] in logging._nameToLevel:
                api_log.setLevel(self.active_config['api_log_level'])
            else:
                raise ConfigurationError(f'Invalid log level: "{self.active_config["api_log_level"]}" not in {list(logging._nameToLevel.keys())}')
        log.addHandler(self.log_storage)
        api_log.addHandler(self.api_log_storage)

    def fetch_all(self) -> None:
        """
        Fetches all data from the connector and updates the CarConnectivity system.

        This method should be overridden by subclasses to implement the specific
        data fetching logic for the connector.
        """

    def startup(self) -> None:
        """
        Starts the connector.

        This method should be overridden by subclasses to implement any necessary
        startup procedures for the connector. If threads are used, they should be started here.
        """

    def shutdown(self) -> None:
        """
        Shuts down the connector.

        This method should be overridden by subclasses to implement any necessary
        cleanup or shutdown procedures for the connector. If threads are used, they should be stopped here.
        """

    def get_version(self) -> str:
        """
        Returns the version of the connector.

        Returns:
            str: The version of the connector.
        """
        raise NotImplementedError("Method get_version() must be implemented by connector")

    def get_type(self) -> str:
        """
        Returns the type of the connector.

        Returns:
            str: The type of the connector.
        """
        raise NotImplementedError("Method get_type() must be implemented by plugin")

    def is_healthy(self) -> bool:
        """
        Returns whether the connector is healthy.

        Returns:
            bool: True if the connector is healthy, False otherwise.
        """
        if self.healthy.enabled and self.healthy.value is not None:
            return self.healthy.value
        return False

    def get_name(self) -> str:
        """
        Returns the user readable name of the connector.
        If not implemented by the connector, fallback is the ID of the connector.

        Returns:
            str: The name of the connector.
        """
        return self.id



"""
This script is an example showing how to retrieve all vehicles from the account using the CarConnectivity library.

Functions:
    main: The main function that sets up argument parsing, logging, and retrieves vehicle data.

Usage:
    python all_cars.py <config> [--tokenstorefile TOKENSTOREFILE] [-v] [--logging-format LOGGING_FORMAT] [--logging-date-format LOGGING_DATE_FORMAT]

Arguments:
    config: Path to the configuration file.
    --tokenstorefile: File to store tokenstore (default: /tmp/tokenstore).
    -v, --verbose: Logging level (verbosity).
    --logging-format: Logging format configured for python logging (default: %(asctime)s:%(levelname)s:%(message)s).
    --logging-date-format: Logging date format configured for python logging (default: %Y-%m-%dT%H:%M:%S%z).

Logging Levels:
    DEBUG, INFO, WARNING, ERROR, CRITICAL

Example:
    python all_cars.py config.json --tokenstorefile /path/to/tokenstore -v
"""
from __future__ import annotations
from typing import TYPE_CHECKING

import argparse
import json
import os
import tempfile
import logging

from carconnectivity import carconnectivity

if TYPE_CHECKING:
    from typing import List

LOG_LEVELS: List[str] = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_LOG_LEVEL = "ERROR"
LOG: logging.Logger = logging.getLogger("carconnectivity-example")


#  pylint: disable=duplicate-code
def main() -> None:
    """ Simple example showing how to retrieve all vehicles from the account """
    parser = argparse.ArgumentParser(
        prog='allCars',
        description='Example retrieving all cars in all configured connectors'
    )
    parser.add_argument('config', help='Path to the configuration file')
    default_temp: str = os.path.join(tempfile.gettempdir(), 'tokenstore')
    parser.add_argument('--tokenstorefile', help=f'file to store tokenstore (default: {default_temp})', default=default_temp)

    logging_group = parser.add_argument_group('Logging')
    logging_group.add_argument('-v', '--verbose', action="append_const", help='Logging level (verbosity)', const=-1,)
    logging_group.add_argument('--logging-format', dest='logging_format', help='Logging format configured for python logging '
                               '(default: %%(asctime)s:%%(module)s:%%(message)s)', default='%(asctime)s:%(levelname)s:%(message)s')
    logging_group.add_argument('--logging-date-format', dest='logging_date_format', help='Logging format configured for python logging '
                               '(default: %%Y-%%m-%%dT%%H:%%M:%%S%%z)', default='%Y-%m-%dT%H:%M:%S%z')

    args = parser.parse_args()

    log_level: int = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
    for adjustment in args.verbose or ():
        log_level = min(len(LOG_LEVELS) - 1, max(log_level + adjustment, 0))

    logging.basicConfig(level=LOG_LEVELS[log_level], format=args.logging_format, datefmt=args.logging_date_format)

    print('#  read CarConnectivity configuration')
    with open(args.config, 'r', encoding='utf-8') as config_file:
        config_dict = json.load(config_file)
        print('#  Login')
        car_connectivity = carconnectivity.CarConnectivity(config=config_dict, tokenstore_file=args.tokenstorefile)
        print('#  getData')
        car_connectivity.get_garage()
        print('#  Shutdown')
        car_connectivity.shutdown()

    print('#  done')


if __name__ == '__main__':
    main()



"""Module for position classes."""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import EnumAttribute, TemperatureAttribute, DateAttribute, BooleanAttribute
from carconnectivity.commands import Commands

if TYPE_CHECKING:
    from typing import Optional

    from carconnectivity.vehicle import GenericVehicle


class Climatization(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent a climatization.
    """
    def __init__(self, vehicle: Optional[GenericVehicle] = None, origin: Optional[Climatization] = None) -> None:
        if origin is not None:
            super().__init__(origin=origin)
            self.commands: Commands = origin.commands
            self.commands.parent = self
            self.state: EnumAttribute = origin.state
            self.state.parent = self
            self.estimated_date_reached: DateAttribute = origin.estimated_date_reached
            self.estimated_date_reached.parent = self
            self.settings: Climatization.Settings = Climatization.Settings(origin=origin.settings)
        else:
            super().__init__(object_id='climatization', parent=vehicle)
            self.commands: Commands = Commands(parent=self)
            self.state: EnumAttribute = EnumAttribute("state", self, value_type=Climatization.ClimatizationState, tags={'carconnectivity'})
            self.estimated_date_reached: DateAttribute = DateAttribute("estimated_date_reached", self, tags={'carconnectivity'})
            self.settings: Climatization.Settings = Climatization.Settings(parent=self)

    class Settings(GenericObject):
        """
        This class represents the settings for car  charging.
        """
        def __init__(self, parent: Optional[GenericObject] = None, origin: Optional[Climatization.Settings] = None) -> None:
            if origin is not None:
                super().__init__(parent=parent, origin=origin)
                self.commands: Commands = origin.commands
                self.commands.parent = self
                self.target_temperature: TemperatureAttribute = origin.target_temperature
                self.target_temperature.parent = self
                self.window_heating: BooleanAttribute = origin.window_heating
                self.window_heating.parent = self
                self.seat_heating: BooleanAttribute = origin.seat_heating
                self.seat_heating.parent = self
                self.climatization_at_unlock: BooleanAttribute = origin.climatization_at_unlock
                self.climatization_at_unlock.parent = self
                self.climatization_without_external_power: BooleanAttribute = origin.climatization_without_external_power
                self.climatization_without_external_power.parent = self
                self.heater_source: EnumAttribute = origin.heater_source
                self.heater_source.parent = self
            else:
                super().__init__(object_id="settings", parent=parent)
                self.commands: Commands = Commands(parent=self)
                self.target_temperature: TemperatureAttribute = TemperatureAttribute("target_temperature", parent=self, tags={'carconnectivity'})
                self.window_heating: BooleanAttribute = BooleanAttribute("window_heating", parent=self, tags={'carconnectivity'})
                self.seat_heating: BooleanAttribute = BooleanAttribute("seat_heating", parent=self, tags={'carconnectivity'})
                self.climatization_at_unlock: BooleanAttribute = BooleanAttribute("climatization_at_unlock", parent=self, tags={'carconnectivity'})
                self.climatization_without_external_power: BooleanAttribute = BooleanAttribute("climatization_without_external_power", parent=self,
                                                                                               tags={'carconnectivity'})
                self.heater_source: EnumAttribute = EnumAttribute("heater_source", parent=self, value_type=Climatization.Settings.HeaterSource,
                                                                  tags={'carconnectivity'})

        class HeaterSource(Enum,):
            """
            Enum representing different sources of a heater.

            Attributes:
                ELECTRIC (str): Represents an electric heater source.
                UNKNOWN (str): Represents an unknown heater source.
            """
            ELECTRIC = 'electric'
            UNKNOWN = 'unknown heater source'

    class ClimatizationState(Enum,):
        """
        Enum representing the state of the climatization system.

        Attributes:
            OFF (str): Climatization is turned off.
            HEATING (str): Climatization is in heating mode.
            COOLING (str): Climatization is in cooling mode.
            VENTILATION (str): Climatization is in ventilation mode.
            INVALID (str): Climatization state is invalid.
            UNKNOWN (str): Climatization state is unknown.
        """
        OFF = 'off'
        HEATING = 'heating'
        COOLING = 'cooling'
        VENTILATION = 'ventilation'
        INVALID = 'invalid'
        UNKNOWN = 'unknown climatization state'



"""Module for car images."""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import ImageAttribute

if TYPE_CHECKING:
    from typing import Optional, Dict
    from carconnectivity.vehicle import GenericVehicle


class Images(GenericObject):  # pylint: disable=too-many-instance-attributes
    """
    A class to represent all vehicle images.
    """
    def __init__(self, vehicle: Optional[GenericVehicle] = None) -> None:
        if vehicle is None:
            raise ValueError('Cannot create images without vehicle')
        super().__init__(object_id='images', parent=vehicle)
        self.images: Dict[str, ImageAttribute] = {}



"""
Module for information about the vehicle software
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import StringAttribute

if TYPE_CHECKING:
    from carconnectivity.vehicle import GenericVehicle


class Software(GenericObject):
    """
    Represents the software of a vehicle.
    """
    def __init__(self, vehicle: GenericVehicle) -> None:
        super().__init__(object_id='software', parent=vehicle)
        self.version = StringAttribute('version', parent=self, tags={'carconnectivity'})



"""Module containing the commandline interface for the carconnectivity package."""
from __future__ import annotations
from typing import TYPE_CHECKING

import sys
import os
import argparse
import logging
import tempfile
import time
import json

from json_minify import json_minify

from carconnectivity import carconnectivity, errors, util
from carconnectivity._version import __version__ as __carconnectivity_version__

if TYPE_CHECKING:
    from typing import Optional

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
DEFAULT_LOG_LEVEL = "ERROR"

LOG: logging.Logger = logging.getLogger("carconnectivity")


class CLI():  # pylint: disable=too-few-public-methods
    """
    Class containing the commandline interface for the carconnectivity package.
    """

    def __init__(self, logger: logging.Logger, name: str, description: str, subversion: Optional[str] = None) -> None:
        self.logger = logger

        self.parser = argparse.ArgumentParser(
            prog=name,
            description=description)
        if subversion is not None:
            version = f'%(prog)s {subversion} (using CarConnectivity {__carconnectivity_version__})'
        else:
            version = f'%(prog)s {__carconnectivity_version__}'
        self.parser.add_argument('--version', action='version', version=version)
        self.parser.add_argument('config', help='Path to the configuration file')

        default_temp = os.path.join(tempfile.gettempdir(), 'carconnectivity.token')
        self.parser.add_argument('--tokenfile', help=f'file to store token (default: {default_temp})', default=default_temp)
        default_cache_temp = os.path.join(tempfile.gettempdir(), 'carconnectivity.cache')
        self.parser.add_argument('--cachefile', help=f'file to store cache (default: {default_cache_temp})', default=default_cache_temp)
        self.parser.add_argument('--healthcheckfile', help='file to store healthcheck data', default=None)

        logging_group = self.parser.add_argument_group('Logging')
        logging_group.add_argument('-v', '--verbose', action="append_const", help='Logging level (verbosity)', const=-1,)
        logging_group.add_argument('--logging-format', dest='logging_format', help='Logging format configured for python logging '
                                   '(default: %%(asctime)s:%%(module)s:%%(message)s)', default='%(asctime)s:%(levelname)s:%(message)s')
        logging_group.add_argument('--logging-date-format', dest='logging_date_format', help='Logging format configured for python logging '
                                   '(default: %%Y-%%m-%%dT%%H:%%M:%%S%%z)', default='%Y-%m-%dT%H:%M:%S%z')
        logging_group.add_argument('--hide-repeated-log', dest='hide_repeated_log', help='Hide repeated log messages from the same module', action='store_true')

    # pylint: disable-next=too-many-statements,too-many-branches,too-many-locals
    def main(self) -> None:  # noqa: C901
        """
        Entry point for the command-line interface.
        """
        args = self.parser.parse_args()
        log_level = LOG_LEVELS.index(DEFAULT_LOG_LEVEL)
        for adjustment in args.verbose or ():
            log_level = min(len(LOG_LEVELS) - 1, max(log_level + adjustment, 0))

        logging.basicConfig(level=LOG_LEVELS[log_level], format=args.logging_format, datefmt=args.logging_date_format)
        if args.hide_repeated_log:
            for handler in logging.root.handlers:
                handler.addFilter(util.DuplicateFilter())

        try:  # pylint: disable=too-many-nested-blocks
            try:
                with open(file=args.config, mode='r', encoding='utf-8') as config_file:
                    try:
                        config_dict = json.loads(json_minify(config_file.read(), strip_space=False))
                        car_connectivity = carconnectivity.CarConnectivity(config=config_dict, tokenstore_file=args.tokenfile, cache_file=args.cachefile)
                        car_connectivity.startup()
                        try:
                            while True:
                                if args.healthcheckfile is not None:
                                    with open(file=args.healthcheckfile, mode='w', encoding='utf-8') as healthcheck_file:
                                        if car_connectivity.is_healthy():
                                            healthcheck_file.write('healthy')
                                        else:
                                            healthcheck_file.write('unhealthy')
                                time.sleep(60)
                        except KeyboardInterrupt:
                            self.logger.info('Keyboard interrupt received, shutting down...')

                            car_connectivity.shutdown()
                    except json.JSONDecodeError as e:
                        self.logger.critical('Could not load configuration file %s (%s)', args.config, e)
                        sys.exit('Could not load configuration file')
            except FileNotFoundError as e:
                self.logger.critical('Could not find configuration file %s (%s)', args.config, e)
                sys.exit('Could not find configuration file')
        except errors.AuthenticationError as e:
            self.logger.critical('There was a problem when authenticating with one or multiple services: %s', e)
            sys.exit('There was a problem when authenticating with one or multiple services')
        except errors.APICompatibilityError as e:
            self.logger.critical('There was a problem when communicating with one or multiple services.'
                                 ' If this problem persists please open a bug report: %s', e)
            sys.exit('There was a problem when communicating with one or multiple services.')
        except errors.RetrievalError as e:
            self.logger.critical('There was a problem when communicating with one or multiple services: %s', e)
            sys.exit('There was a problem when communicating with one or multiple services.')
        except errors.ConfigurationError as e:
            self.logger.critical('There was a problem with the configuration: %s', e)
            sys.exit('There was a problem with the configuration')
        except KeyboardInterrupt:
            sys.exit("killed")


def main() -> None:
    """
    Entry point for the car connectivity application.

    This function initializes and starts the command-line interface (CLI) for the
    car connectivity application using the specified logger and application name.
    """
    cli: CLI = CLI(logger=LOG, name='carconnectivity', description='Commandline Interface to interact with Car Services of various brands')
    cli.main()



"""
Module containing the Garage class.

This module defines the Garage class, which represents a garage that can hold multiple vehicles.
The Garage class provides methods to add, replace, remove, and retrieve vehicles, as well as list all vehicles and their VINs.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject
from carconnectivity.vehicle import GenericVehicle

if TYPE_CHECKING:
    from typing import List, Dict, Optional


class Garage(GenericObject):
    """
    A class to represent a garage that can hold multiple vehicles.
    """
    def __init__(self, parent) -> None:
        super().__init__(object_id='garage', parent=parent)
        self.delay_notifications = True
        self._vehicles: Dict[str, GenericVehicle] = {}
        self.enabled = True
        self.delay_notifications = False

    def add_vehicle(self, vehicle_id: str, vehicle: GenericVehicle) -> None:
        """
        Adds a vehicle to the garage.

        Args:
            vehicle_id (str): The unique identifier of the vehicle.
            vehicle (GenericVehicle): The vehicle object to be added.

        Returns:
            None
        """
        if vehicle_id.upper() in self._vehicles:
            raise ValueError(f'Vehicle with ID {vehicle_id.upper()} already exists in the garage.')
        self._vehicles[vehicle_id.upper()] = vehicle

    def replace_vehicle(self, vehicle_id: str, vehicle: GenericVehicle) -> None:
        """
        Replaces a vehicle in the garage.

        Args:
            vehicle_id (str): The unique identifier of the vehicle.
            vehicle (GenericVehicle): The vehicle object to be added.

        Returns:
            None
        """
        self._vehicles[vehicle_id.upper()] = vehicle

    def remove_vehicle(self, vehicle_id: str) -> None:
        """
        Remove a vehicle from the garage by its vehicle ID.

        Args:
            vehicle_id (str): The ID of the vehicle to be removed.

        Returns:
            None
        """
        if vehicle_id.upper() in self._vehicles:
            del self._vehicles[vehicle_id.upper()]

    def get_vehicle(self, vehicle_id: str) -> Optional[GenericVehicle]:
        """
        Retrieve a vehicle from the garage by its vehicle ID.

        Args:
            vehicle_id (str): The unique identifier of the vehicle to retrieve.

        Returns:
            GenericVehicle | None: The vehicle object if found, otherwise None.
        """
        return self._vehicles.get(vehicle_id.upper())

    def list_vehicles(self) -> List[GenericVehicle]:
        """
        Returns a list of all vehicles in the garage.

        Returns:
            list[GenericVehicle]: A list containing all the vehicles.
        """
        return list(self._vehicles.values())

    def list_vehicle_vins(self) -> List[str]:
        """
        Returns a list of all vehicle vins of vehicles in the garage.

        Returns:
            list[str]: A list containing all the vehicle vins.
        """
        return list(self._vehicles.keys())



"""
This module defines the collection of connectors.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject

if TYPE_CHECKING:
    from typing import Dict

    from carconnectivity.carconnectivity import CarConnectivity

    from carconnectivity_connectors.base.connector import BaseConnector


class Connectors(GenericObject):
    """
    Represents the collection of connectors.
    """
    def __init__(self, car_connectivity: CarConnectivity) -> None:
        super().__init__(object_id='connectors', parent=car_connectivity)
        self.connectors: Dict[str, BaseConnector] = {}



"""Module containing custom json encoders for the carconnectivity package."""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

# pylint: disable=duplicate-code
SUPPORT_IMAGES = False
try:
    from PIL import Image
    SUPPORT_IMAGES = True
except ImportError:
    pass

SUPPORT_ASCII_IMAGES = False
try:
    import ascii_magic
    import shutil
    SUPPORT_ASCII_IMAGES = True
except ImportError:
    pass
# pylint: enable=duplicate-code

if TYPE_CHECKING:
    from typing import Optional, Tuple

if SUPPORT_IMAGES and SUPPORT_ASCII_IMAGES:
    class ASCIIModes(Enum):
        """
        Enum class representing different ASCII modes.

        Attributes:
            TERMINAL (str): Mode for terminal output.
            ASCII (str): Mode for plain ASCII output.
            HTML (str): Mode for HTML output.
        """
        TERMINAL = 'TERMINAL'
        ASCII = 'ASCII'
        HTML = 'HTML'

    def image_to_ASCII_art(img: Image.Image, columns: int = 0, mode=ASCIIModes.TERMINAL) -> str:  # pylint: disable=invalid-name
        """
        Converts an image to ASCII art.

        Args:
            img (Image): The input image to be converted.
            columns (int, optional): The number of columns for the ASCII art. Defaults to 0, which uses the terminal width.
            mode (ASCIIModes, optional): The mode for the ASCII art output. Can be ASCIIModes.TERMINAL, ASCIIModes.ASCII, or ASCIIModes.HTML.
            Defaults to ASCIIModes.TERMINAL.

        Returns:
            str: The resulting ASCII art as a string.
        """
        bbox: Optional[Tuple[int, int, int, int]] = img.getbbox()

        # Crop the image to the contents of the bounding box
        image: Image.Image = img.crop(bbox)

        # Determine the width and height of the cropped image
        (width, height) = image.size

        # Create a new image object for the output image
        cropped_image: Image.Image = Image.new("RGBA", (width, height), (0, 0, 0, 0))  # pyright: ignore[reportPossiblyUnboundVariable]

        # Paste the cropped image onto the new image
        cropped_image.paste(image, (0, 0))

        if columns == 0:
            columns = shutil.get_terminal_size()[0]  # pyright: ignore[reportPossiblyUnboundVariable]

        if mode == ASCIIModes.ASCII:
            return ascii_magic.from_pillow_image(cropped_image).to_ascii(columns=columns, monochrome=True)  # pyright: ignore[reportPossiblyUnboundVariable]
        if mode == ASCIIModes.HTML:
            return ascii_magic.from_pillow_image(cropped_image).to_html(columns=columns)  # pyright: ignore[reportPossiblyUnboundVariable]
        return ascii_magic.from_pillow_image(cropped_image).to_ascii(columns=columns)  # pyright: ignore[reportPossiblyUnboundVariable]



"""
This module defines the collection of plugins.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.objects import GenericObject

if TYPE_CHECKING:
    from typing import Dict

    from carconnectivity.carconnectivity import CarConnectivity

    from carconnectivity_plugins.base.plugin import BasePlugin


class Plugins(GenericObject):
    """
    Represents the collection of plugins.
    """
    def __init__(self, car_connectivity: CarConnectivity) -> None:
        super().__init__(object_id='plugins', parent=car_connectivity)
        self.plugins: Dict[str, BasePlugin] = {}



"""
This module defines classes related to vehicle drives.

It includes the Drives class to manage multiple drives of a vehicle,
and the GenericDrive class along with its subclasses ElectricDrive
and CombustionDrive to represent different types of vehicle drives.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from enum import Enum

from carconnectivity.objects import GenericObject
from carconnectivity.attributes import RangeAttribute, LevelAttribute, EnumAttribute
from carconnectivity.units import Length
from carconnectivity.battery import Battery

if TYPE_CHECKING:
    from typing import Dict
    from carconnectivity.vehicle import GenericVehicle


class Drives(GenericObject):
    """
    Represents the drives of a vehicle.
    """
    def __init__(self, vehicle: GenericVehicle) -> None:
        super().__init__(object_id='drives', parent=vehicle)
        self.total_range: RangeAttribute = RangeAttribute(name="total_range", parent=self, value=None, unit=Length.UNKNOWN, minimum=0, tags={'carconnectivity'})
        self.drives: Dict[str, GenericDrive] = {}

    def add_drive(self, drive: GenericDrive) -> None:
        """
        Adds a drive to the drives of the vehicle.

        Parameters:
        -----------
        drive : GenericDrive
            The drive to add.
        """
        self.drives[drive.id] = drive


class GenericDrive(GenericObject):
    """
    A class to represent a generic drive.
    """
    def __init__(self, drive_id: str, drives: Drives) -> None:
        super().__init__(object_id=drive_id, parent=drives)
        self.type: EnumAttribute = EnumAttribute(name="type", parent=self, value=None, tags={'carconnectivity'})
        self.range: RangeAttribute = RangeAttribute(name="range", parent=self, value=None, unit=Length.UNKNOWN, minimum=0, tags={'carconnectivity'})
        self.level: LevelAttribute = LevelAttribute(name="level", parent=self, value=None, minimum=0, tags={'carconnectivity'})
        self.enabled = True

    # pylint: disable=duplicate-code
    class Type(Enum):
        """
        Enum representing different types of drives.
        """
        ELECTRIC = 'electric'
        FUEL = 'fuel'
        GASOLINE = 'gasoline'
        PETROL = 'petrol'
        DIESEL = 'diesel'
        CNG = 'cng'
        LPG = 'lpg'
        INVALID = 'invalid'
        UNKNOWN = 'unknown drive type'
    # pylint: enable=duplicate-code


class ElectricDrive(GenericDrive):
    """
    Represents an electric drive.
    """
    def __init__(self, drive_id: str, drives: Drives) -> None:
        super().__init__(drive_id=drive_id, drives=drives)
        self.battery: Battery = Battery(drive=self)


class CombustionDrive(GenericDrive):
    """
    Represents an combustion drive.
    """
    def __init__(self, drive_id: str, drives: Drives) -> None:
        super().__init__(drive_id=drive_id, drives=drives)


class DieselDrive(CombustionDrive):
    """
    Represents a diesel combustion drive.
    """
    def __init__(self, drive_id: str, drives: Drives) -> None:
        super().__init__(drive_id=drive_id, drives=drives)
        self.adblue_range: RangeAttribute = RangeAttribute(name="adblue_range", parent=self, value=None, unit=Length.UNKNOWN, minimum=0,
                                                           tags={'carconnectivity'})
        self.adblue_level: LevelAttribute = LevelAttribute(name="adblue_level", parent=self, value=None, minimum=0, tags={'carconnectivity'})
